pub mod ast;

use pest::Parser;
use pest::iterators::Pair;
use pest_derive::Parser;
use thiserror::Error;

#[derive(Parser)]
#[grammar = "solace.pest"]
pub struct SolaceParser;

#[derive(Error, Debug)]
pub enum ParseError {
    #[error("Parse error: {0}")]
    PestError(#[from] pest::error::Error<Rule>),

    #[error("Unexpected rule: {0:?}")]
    UnexpectedRule(Rule),

    #[error("Statement not implemented: {0:?}")]
    UnknownStatement(Rule),

    #[error("Unknown keyword, literal or expression: {0:?}")]
    UnknownPrimary(Rule),

    #[error("Left side of assignment must be an identifier: {0:?}")]
    InvalidLeftHand(ast::Expr),

    #[error("Unexpected infix operator: {0:?}")]
    UnknownInfixOperator(Rule),

    #[error("Unexpected prefix operator: {0:?}")]
    UnknownPrefixOperator(Rule),

    #[error("Unexpected postfix operator: {0:?}")]
    UnknownPostfixOperator(Rule),

    #[error("Expected type annotation or type expression, got: {0:?}")]
    ShouldBeType(Rule),

    #[error("Unexpected return type: \"{0}\" ({1:?})")]
    UnknownReturnType(String, Rule),

    #[error("Unexpected base type: {0:?}")]
    UnknownTypePart(Rule),

    #[error("Invalid Number: {0:?}")]
    InvalidNumber(Rule),

    #[error("No base type found")]
    MissingBaseType(),

    #[error("Function body required: {0}")]
    MissingFunctionBody(String),
}

pub fn parse(input: &str, debug: bool) -> Result<ast::Program, ParseError> {
    if debug {
        let mut debug_pairs = SolaceParser::parse(Rule::program, input)?;
        print_parse_tree(debug_pairs.next().unwrap(), 0);
    }

    let pairs = SolaceParser::parse(Rule::program, input)?;
    let program = ast::Program::from_pairs(pairs)?;

    Ok(program)
}

fn print_parse_tree(pair: Pair<Rule>, indent: usize) {
    let indent_str = "  ".repeat(indent);
    println!("{}{:?}: \"{}\"", indent_str, pair.as_rule(), pair.as_str());

    for inner_pair in pair.into_inner() {
        print_parse_tree(inner_pair, indent + 1)
    }
}
